##
## Green/blue and canary deployments
##

Container Images are based on NodeJS and have been build to support both AMD64 and ARM64 architectures.

#1. Step 1 - Create NS

kubectl create ns noco

#2. Step 2 - Deploy green deployment

pascal@rpb8c224n01:~/green-blue-deployments$ kubectl apply -f noco-green-deploy.yaml -n noco
deployment.apps/noco-1.0 created

--> verify


pascal@rpb8c224n01:~/green-blue-deployments$ kubectl get pods -n noco -o wide
NAME                        READY   STATUS    RESTARTS   AGE   IP           NODE          NOMINATED NODE   READINESS GATES
noco-1.0-569b7db995-ct58w   1/1     Running   0          5s    10.0.1.204   rpb8c224n03   <none>           <none>
noco-1.0-569b7db995-kpjr6   1/1     Running   0          5s    10.0.2.92    rpb8c224n04   <none>           <none>
noco-1.0-569b7db995-lthr7   1/1     Running   0          5s    10.0.1.230   rpb8c224n03   <none>           <none>

--> check if 1.0 (green POD) responds on port :3000

pascal@rpb8c224n01:~/green-blue-deployments$ curl 10.0.1.204:3000
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta http-equiv="refresh" content="5" >
    <title>NodeJS Color Container App</title>
</head>
<body>
        <div style="text-align:center;">
        <img width="300" src="images/cc.svg">

        <H2>Welcome to the green container</H2>
        <P>
        This container is running on <I>noco-1.0-569b7db995-ct58w</I> (<I>10.0.1.204</I>)
        <P>
        Started on: <I>2022-10-29 13:52:36</I>
        <P>
        Running for <I>119.685</I> seconds.
        <P>
        You are visitor nr <I>1</I> connecting from: <I>::ffff:10.0.0.65</I>
        <P>
        <H3>Noco version: 0.9</H3>
</body>
</html>

--> Green container, ignore the Noco version; they are all the same. The color is green for 1.0 and blue for 1.1, that's what matters

NOTE: In the POD spec for the green deployment we put two labels on each POD deployed using pod.spec.template ->

 template:
    metadata:
      creationTimestamp: null
      labels:
        app: noco
        version: "1.0"
    spec:
      containers:
      - image: quay.io/pamvdam/noco:v9
        name: noco
        env:
          - name: CCOLOR
            value: green
        ports:
        - containerPort: 3000


#3. Step 3 - Deploy blue deployment

pascal@rpb8c224n01:~/green-blue-deployments$ kubectl apply -f noco-blue-deploy.yaml -n noco
deployment.apps/noco-1.1 created
pascal@rpb8c224n01:~/green-blue-deployments$ kubectl get pods -o wide -n noco
NAME                        READY   STATUS    RESTARTS   AGE     IP           NODE          NOMINATED NODE   READINESS GATES
noco-1.0-569b7db995-ct58w   1/1     Running   0          7m42s   10.0.1.204   rpb8c224n03   <none>           <none>
noco-1.0-569b7db995-kpjr6   1/1     Running   0          7m42s   10.0.2.92    rpb8c224n04   <none>           <none>
noco-1.0-569b7db995-lthr7   1/1     Running   0          7m42s   10.0.1.230   rpb8c224n03   <none>           <none>
noco-1.1-676f95987c-7b6xl   1/1     Running   0          11s     10.0.1.39    rpb8c224n03   <none>           <none>
noco-1.1-676f95987c-9flh7   1/1     Running   0          11s     10.0.2.217   rpb8c224n04   <none>           <none>
noco-1.1-676f95987c-jqv77   1/1     Running   0          11s     10.0.1.93    rpb8c224n03   <none>           <none>
pascal@rpb8c224n01:~/green-blue-deployments$ curl 10.0.1.39:3000
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta http-equiv="refresh" content="5" >
    <title>NodeJS Color Container App</title>
</head>
<body>
        <div style="text-align:center;">
        <img width="300" src="images/cc.svg">

        <H2>Welcome to the blue container</H2>
        <P>
        This container is running on <I>noco-1.1-676f95987c-7b6xl</I> (<I>10.0.1.39</I>)
        <P>
        Started on: <I>2022-10-29 14:00:07</I>
        <P>
        Running for <I>17.713</I> seconds.
        <P>
        You are visitor nr <I>1</I> connecting from: <I>::ffff:10.0.0.65</I>
        <P>
        <H3>Noco version: 0.9</H3>
</body>
</html>


--> Container is blue; so this is 1.1

--> If a service selects PODs with version=1.0 -> green PODs will be targeted

pascal@rpb8c224n01:~/green-blue-deployments$ kubectl get pods -n noco -l version=1.0
NAME                        READY   STATUS    RESTARTS   AGE
noco-1.0-569b7db995-ct58w   1/1     Running   0          8m7s
noco-1.0-569b7db995-kpjr6   1/1     Running   0          8m7s
noco-1.0-569b7db995-lthr7   1/1     Running   0          8m7s

-- If a service selects PODs with version=1.1 -> blue PODs will be targeted

pascal@rpb8c224n01:~/green-blue-deployments$ kubectl get pods -n noco -l version=1.1
NAME                        READY   STATUS    RESTARTS   AGE
noco-1.1-676f95987c-7b6xl   1/1     Running   0          39s
noco-1.1-676f95987c-9flh7   1/1     Running   0          39s
noco-1.1-676f95987c-jqv77   1/1     Running   0          39s

#4. Step 4 - create the service to address both versions (actually a 50/50 canary):

pascal@rpb8c224n01:~/green-blue-deployments$ kubectl get svc -n noco noco
NAME   TYPE        CLUSTER-IP      EXTERNAL-IP   PORT(S)    AGE
noco   ClusterIP   10.111.56.149   <none>        3000/TCP   0s
pascal@rpb8c224n01:~/green-blue-deployments$ kubectl describe  svc -n noco noco
Name:              noco
Namespace:         noco
Labels:            app=noco
Annotations:       <none>
Selector:          app=noco,version=1.0
Type:              ClusterIP
IP Family Policy:  SingleStack
IP Families:       IPv4
IP:                10.111.56.149
IPs:               10.111.56.149
Port:              <unset>  3000/TCP
TargetPort:        3000/TCP
Endpoints:         10.0.1.204:3000,10.0.1.230:3000,10.0.2.92:3000
Session Affinity:  None
Events:            <none>
pascal@rpb8c224n01:~/green-blue-deployments$ kubectl get pods -o wide -n noco -l app=noco,version=1.0
NAME                        READY   STATUS    RESTARTS   AGE   IP           NODE          NOMINATED NODE   READINESS GATES
noco-1.0-569b7db995-ct58w   1/1     Running   0          13m   10.0.1.204   rpb8c224n03   <none>           <none>
noco-1.0-569b7db995-kpjr6   1/1     Running   0          13m   10.0.2.92    rpb8c224n04   <none>           <none>
noco-1.0-569b7db995-lthr7   1/1     Running   0          13m   10.0.1.230   rpb8c224n03   <none>           <none>
pascal@rpb8c224n01:~/green-blue-deployments$ 

So.. with app=noco AND version=1.0 is we get the green POD release (1.0).

Now let's operate the service switch:

kubectl patch service noco -p '{"spec":{"selector":{"version": "1.1"}}}

pascal@rpb8c224n01:~/green-blue-deployments$ kubectl patch service noco -p '{"spec":{"selector":{"version": "1.1"}}}'  -n noco
service/noco patched
pascal@rpb8c224n01:~/green-blue-deployments$ kubectl describe  svc -n noco noco
Name:              noco
Namespace:         noco
Labels:            app=noco
Annotations:       <none>
Selector:          app=noco,version=1.1
Type:              ClusterIP
IP Family Policy:  SingleStack
IP Families:       IPv4
IP:                10.111.56.149
IPs:               10.111.56.149
Port:              <unset>  3000/TCP
TargetPort:        3000/TCP
Endpoints:         10.0.1.39:3000,10.0.1.93:3000,10.0.2.217:3000
Session Affinity:  None
Events:            <none>
pascal@rpb8c224n01:~/green-blue-deployments$ kubectl get pods -o wide -n noco -l app=noco,version=1.1
NAME                        READY   STATUS    RESTARTS   AGE     IP           NODE          NOMINATED NODE   READINESS GATES
noco-1.1-676f95987c-7b6xl   1/1     Running   0          8m20s   10.0.1.39    rpb8c224n03   <none>           <none>
noco-1.1-676f95987c-9flh7   1/1     Running   0          8m20s   10.0.2.217   rpb8c224n04   <none>           <none>
noco-1.1-676f95987c-jqv77   1/1     Running   0          8m20s   10.0.1.93    rpb8c224n03   <none>           <none>
pascal@rpb8c224n01:~/green-blue-deployments$ curl 10.111.56.149:3000
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta http-equiv="refresh" content="5" >
    <title>NodeJS Color Container App</title>
</head>
<body>
        <div style="text-align:center;">
        <img width="300" src="images/cc.svg">

        <H2>Welcome to the blue container</H2>
        <P>
        This container is running on <I>noco-1.1-676f95987c-jqv77</I> (<I>10.0.1.93</I>)
        <P>
        Started on: <I>2022-10-29 14:00:07</I>
        <P>
        Running for <I>517.76</I> seconds.
        <P>
        You are visitor nr <I>1</I> connecting from: <I>::ffff:10.0.0.65</I>
        <P>
        <H3>Noco version: 0.9</H3>
</body>
</html>


--> So after the patch we target the blue PODs release

If we regret it, we rollback using a patch on the svc for version 1.0 like this:


pascal@rpb8c224n01:~/green-blue-deployments$ kubectl patch service noco -p '{"spec":{"selector":{"version": "1.0"}}}'  -n noco
service/noco patched
pascal@rpb8c224n01:~/green-blue-deployments$ curl 10.111.56.149:3000
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta http-equiv="refresh" content="5" >
    <title>NodeJS Color Container App</title>
</head>
<body>
        <div style="text-align:center;">
        <img width="300" src="images/cc.svg">

        <H2>Welcome to the green container</H2>
        <P>
        This container is running on <I>noco-1.0-569b7db995-ct58w</I> (<I>10.0.1.204</I>)
        <P>
        Started on: <I>2022-10-29 13:52:36</I>
        <P>
        Running for <I>983.297</I> seconds.
        <P>
        You are visitor nr <I>2</I> connecting from: <I>::ffff:10.0.0.65</I>
        <P>
        <H3>Noco version: 0.9</H3>
</body>
</html>

--> E.g. back to business as usual

Bonus -> removing the version selector from the noco service turns it into a 50/50 canary release:

kubectl patch service noco --type=json -p='[{"op": "remove", "path": "/spec/selector/version"}]'

pascal@rpb8c224n01:~/green-blue-deployments$ kubectl patch service noco --type=json -p='[{"op": "remove", "path": "/spec/selector/version"}]' -n noco
service/noco patched
pascal@rpb8c224n01:~/green-blue-deployments$ kubectl describe svc noco -n noco
Name:              noco
Namespace:         noco
Labels:            app=noco
Annotations:       <none>
Selector:          app=noco
Type:              ClusterIP
IP Family Policy:  SingleStack
IP Families:       IPv4
IP:                10.111.56.149
IPs:               10.111.56.149
Port:              <unset>  3000/TCP
TargetPort:        3000/TCP
Endpoints:         10.0.1.204:3000,10.0.1.230:3000,10.0.1.39:3000 + 3 more...
Session Affinity:  None
Events:            <none>
pascal@rpb8c224n01:~/green-blue-deployments$ kubectl describe ep noco -n noco
Name:         noco
Namespace:    noco
Labels:       app=noco
Annotations:  endpoints.kubernetes.io/last-change-trigger-time: 2022-10-29T14:00:07Z
Subsets:
  Addresses:          10.0.1.204,10.0.1.230,10.0.1.39,10.0.1.93,10.0.2.217,10.0.2.92
  NotReadyAddresses:  <none>
  Ports:
    Name     Port  Protocol
    ----     ----  --------
    <unset>  3000  TCP

Events:  <none>


